from django.utils.encoding import force_text
from django.conf import settings

from ..util import merge_dicts

import os
import os.path



##############################################################
###   Static File Providers

class BaseProvider(object):
    '''
    Abstract base provider class.  An instance is tied to a template at runtime.
    '''

    default_options = {
        'group': 'styles',
        'enabled': True,
    }

    def __init__(self, app_config, template_file, options):
        # In production, this constructor is called only once per template per server run
        self.app_config = app_config
        self.subdir, self.template_file = os.path.split(template_file)
        self.subdir_parts = os.path.normpath(self.subdir).split(os.path.sep)
        self.template_name, _ = os.path.splitext(self.template_file)
        self.options = merge_dicts(self.default_options, options)     # combined options dictionary

    @property
    def group(self):
        return self.options['group']

    def start(self, provider_run, data):
        '''
        Called on the *main* template's provider list as the run starts.
        Initialize values in the data dictionary here.
        '''
        pass

    def provide(self, provider_run, data):
        '''Called on *each* template's provider list in the chain - use provider_run.write() for content'''
        pass

    def finish(self, provider_run, data):
        '''
        Called on the *main* template's provider list as the run finishes
        Finalize values in the data dictionary here.
        '''
        pass



from django.apps import apps
try:
    from django.urls import re_path              # Django 2.x
except ImportError:
    from django.conf.urls import url as re_path  # Django 1.x
from django.views.static import serve
from .router import route_request
from .registry import get_dmp_apps
from .util import DMP_OPTIONS
import os, os.path


#########################################################
###   The default DMP url patterns
###
###   FYI, even though the valid python identifier is [_A-Za-z][_a-zA-Z0-9]*,
###   I'm simplifying it to [_a-zA-Z0-9]+ because it works for our purposes


# start with the DMP web files - for development time
# at production, serve this directly with Nginx/IIS/etc. instead
# there is no "if debug mode" statement here because the web server will serve the file at production before urls.py happens,
# but if this deployment step isn't done right, it will still work through this link
urlpatterns = [
    re_path(r'^django_mako_plus/(?P<path>[^/]+)', serve, { 'document_root': os.path.join(apps.get_app_config('django_mako_plus').path, 'webroot') }, name='DMP webroot (for devel)'),
]

# app-specific patterns for each DMP-enabled app
for config in get_dmp_apps():
    # these are in order of specificity, with the most specific ones at the top
    urlpatterns.extend([
        # /app/page.function/urlparams
        re_path(r'^{}/(?P<dmp_page>[_a-zA-Z0-9\-]+)\.(?P<dmp_function>[_a-zA-Z0-9\.\-]+)/?(?P<dmp_urlparams>.*?)/?$'.format(config.name), route_request, { 'dmp_app': config.name }, name='DMP /app/page.function'),
        # /app/page/urlparams
        re_path(r'^{}/(?P<dmp_page>[_a-zA-Z0-9\-]+)/?(?P<dmp_urlparams>.*?)/?$'.format(config.name), route_request, { 'dmp_app': config.name }, name='DMP /app/page'),
        # /app
        re_path(r'^{}/?$'.format(config.name), route_request, { 'dmp_app': config.name }, name='DMP /app'),
    ])

# if we have a default app, take over short urls
if DMP_OPTIONS['DEFAULT_APP']:
    urlpatterns.extend([
        # /page.function/urlparams
        re_path(r'^(?P<dmp_page>[_a-zA-Z0-9\-]+)\.(?P<dmp_function>[_a-zA-Z0-9\.\-]*)/?(?P<dmp_urlparams>.*?)/?$', route_request, name='DMP /page.function'),
        # /page/urlparams
        re_path(r'^(?P<dmp_page>[_a-zA-Z0-9\-]+)/?(?P<dmp_urlparams>.*?)/?$', route_request, name='DMP /page'),
        # / with nothing else
        re_path(r'^$', route_request, name='DMP /'),
    ])



# this dict of options is merged with the current project's TEMPLATES entry for DMP
DEFAULT_OPTIONS = {
    # the default app and page to render in Mako when the url is too short
    # if None (no default app), DMP will not capture short URLs
    'DEFAULT_APP': 'homepage',
    'DEFAULT_PAGE': 'index',

    # how to discover and register apps
    # 'default': register all apps that are subfolders of BASE_DIR - e.g. your custom apps
    # 'none': no auto-registration; instead, register apps with `django_mako_plus.register_dmp_app()`
    # 'all': register all apps, even when pip-installed or django apps
    #     Note that 'all' should not normally be used because it makes DMP route *every* app
    #     in your system, including third party apps.  For example, DMP will route /admin/,
    #     which should normally be routed the normal Django way.
    'APP_DISCOVERY': 'default',

    # functions to automatically add variables to the params/context before templates are rendered
    'CONTEXT_PROCESSORS': [
        'django.template.context_processors.static',            # adds "STATIC_URL" from settings.py
        'django.template.context_processors.debug',             # adds debug and sql_queries
        'django.template.context_processors.request',           # adds "request" object
        'django.contrib.auth.context_processors.auth',          # adds "user" and "perms" objects
        'django.contrib.messages.context_processors.messages',  # adds messages from the messages framework
        'django_mako_plus.context_processors.settings',         # adds "settings" dictionary
    ],

    # identifies where the Mako template cache will be stored, relative to each template directory
    'TEMPLATES_CACHE_DIR': '__dmpcache__',

    # the default encoding of template files
    'DEFAULT_TEMPLATE_ENCODING': 'utf-8',

    # imports for every template
    'DEFAULT_TEMPLATE_IMPORTS': [
        # alternative syntax blocks within your Mako templates
        # 'from django_mako_plus import django_syntax, jinja2_syntax, alternate_syntax',

        # the next two lines are just examples of including common imports in templates
        # 'from datetime import datetime',
        # 'import os, os.path, re, json',
    ],

    # whether to send the custom DMP signals -- set to False for a slight speed-up in router processing
    # determines whether DMP will send its custom signals during the process
    'SIGNALS': False,

    # static file providers
    'CONTENT_PROVIDERS': [
        { 'provider': 'django_mako_plus.JsContextProvider' },     # adds JS context - this should normally be listed first
        { 'provider': 'django_mako_plus.CssLinkProvider' },       # generates links for app/styles/template.css
        { 'provider': 'django_mako_plus.JsLinkProvider' },        # generates links for app/scripts/template.js
        # { 'provider': 'django_mako_plus.CompileScssProvider' },   # autocompiles Scss files
        # { 'provider': 'django_mako_plus.CompileLessProvider' },   # autocompiles Less files
    ],

    # webpack providers
    'WEBPACK_PROVIDERS': [
        { 'provider': 'django_mako_plus.JsLinkProvider' },
        { 'provider': 'django_mako_plus.CssLinkProvider' },
    ],

    # additional template dirs to search
    'TEMPLATES_DIRS': [
        # '/var/somewhere/templates/',
    ],
}



from django.core.management.commands.makemessages import Command as MakeMessagesCommand
from django.template.exceptions import TemplateSyntaxError

from django_mako_plus.registry import get_dmp_apps
from django_mako_plus.convenience import get_template_for_path
from django_mako_plus.management.mixins import DMPCommandMixIn

import os
import os.path


class Command(DMPCommandMixIn, MakeMessagesCommand):
    help = (
        "Makes messages for Mako templates.  The native makemessages in Django doesn't understand "
        "Mako files, so this command compiles all your templates so it can find the compiled .py versions of the templates."
    )

    SEARCH_DIRS = [
        os.path.join('{app_path}', 'templates')
    ]

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--extra-gettext-option',
            default=[],
            dest='extra_gettext_option',
            action='append',
            help="Add an additional option to be passed to gettext. Ex: --extra-gettext-option='--keyword=mytrans'"
        )
        parser.add_argument(
            '--ignore-template-errors',
            action='store_true',
            dest='ignore_template_errors',
            default=False,
            help='Ignore any template errors raised when compiling Mako templates'
        )


    def handle(self, *args, **options):
        self.options = options

        # go through each dmp_enabled app and compile its mako templates
        for app_config in get_dmp_apps():
            self.compile_mako_files(app_config)

        # add any extra xgettext_options (the regular makemessages doesn't do this, and I need to include other aliases like _(), _z(), etc.
        for opt in options.get('extra_gettext_option', []):
           self.xgettext_options.append(opt)

        # call the superclass command
        return MakeMessagesCommand.handle(self, *args, **options)


    def compile_mako_files(self, app_config):
        '''Compiles the Mako templates within the apps of this system'''
        # go through the files in the templates, scripts, and styles directories
        for subdir_name in self.SEARCH_DIRS:
            subdir = subdir_name.format(
                app_path=app_config.path,
                app_name=app_config.name,
            )
            if os.path.exists(subdir):
                for filename in os.listdir(subdir):
                    _, ext = os.path.splitext(filename)
                    if ext.lower() in ( '.htm', '.html', '.mako' ):
                        # create the template object, which creates the compiled .py file
                        filepath = os.path.join(subdir, filename)
                        self.message(filepath, 2)
                        try:
                            get_template_for_path(filepath)
                        except TemplateSyntaxError:
                            if not self.options.get('ignore_template_errors'):
                                raise



from urllib.parse import unquote

# try to import MiddlewareMixIn (Django 1.10+)
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    # create a dummy MiddlewareMixin if older Django
    MiddlewareMixin = object

from .registry import ensure_dmp_app
from .router import ClassBasedRouter, get_router
from .util import DMP_OPTIONS, URLParamList, get_dmp_instance


##########################################################
###   Middleware the prepares the request for
###   use with the controller.


class RequestInitMiddleware(MiddlewareMixin):
    '''
    Adds several fields to the request that our controller needs.
    Projects can customize the variables with view middleware that runs after this class.

    This class must be included in settings.py -> MIDDLEWARE.
    '''
    def process_request(self, request):
        '''
        Adds stubs for the DMP custom variables to the request object.

        We do this early in the middleware process because process_view() below gets called
        *after* URL resolution.  If an exception occurs before that (a 404, a middleware exception),
        the variables won't be put on the request object.  Therefore, we put stubs on it at the
        earliest possible place so they exist even if we don't get to process_view().
        '''
        request.dmp = INITIAL_ROUTING_DATA


    def process_view(self, request, view_func, view_args, view_kwargs):
        '''
        Adds the following to the request object:

            request.dmp.app            The Django application (such as "homepage"), as a string.
            request.dmp.page           The view module (such as "index" for index.py), as a string.
            request.dmp.function       The function within the view module to be called (usually "process_request"),
                                       as a string.
            request.dmp.module         The module path in Python terms (such as homepage.views.index), as a string.
            request.dmp.class_obj      The class object, if a class-based view.
            request.dmp.function_obj   The view callable (function, method, etc.) to be called by the router.
            request.dmp.urlparams      A list of the remaining url parts, as a list of strings.
                                       View function parameter conversion uses the values in this list.

        As view middleware, this function runs just before the router.route_request is called.
        '''
        # ensure this hasn't run twice (such as having it in the middleware twice)
        if request.dmp is not INITIAL_ROUTING_DATA:
            return

        # create the RoutingData object
        request.dmp = RoutingData(
            request,
            view_kwargs.pop('dmp_app', None) or DMP_OPTIONS['DEFAULT_APP'],
            view_kwargs.pop('dmp_page', None) or DMP_OPTIONS['DEFAULT_PAGE'],
            view_kwargs.pop('dmp_function', None),
            view_kwargs.pop('dmp_urlparams', '').strip(),
        )


class RoutingData(object):
    '''
    The routing data for DMP.

    An default object is set at the beginning of the request.
    Then after the url pattern is matched, an object with the
    match data is created.
    '''
    def __init__(self, request=None, app=None, page=None, function=None, urlparams=None):
        '''These variables are set by the process_view method above'''
        self.request = request
        # period and dash cannot be in python names, but we allow dash in app, dash in page, and dash/period in function
        self.app = app.replace('-', '_') if app is not None else None
        self.page = page.replace('-', '_') if page is not None else None
        if function:
            self.function = function.replace('.', '_').replace('-', '_')
            fallback_template = '{}.{}.html'.format(page, function)
        else:
            self.function = 'process_request'
            fallback_template = '{}.html'.format(page)

        # set the module and function
        # the return of get_router_function might be a function, a class-based view, or a template
        if self.app is not None and self.page is not None:
            self.module = '.'.join([ self.app, 'views', self.page ])
            self.class_obj = None
            self.function_obj = get_router(self.module, self.function, self.app, fallback_template)
            if isinstance(self.function_obj, ClassBasedRouter):
                self.class_obj = self.function_obj
                self.function = request.method.lower()
        else:
            self.module = None
            self.class_obj = None
            self.function_obj = None

        # parse the urlparams
        # note that I'm not using unquote_plus because the + switches to a space *after* the question mark (in the regular parameters)
        # in the normal url, spaces should be quoted with %20.  Thanks Rosie for the tip.
        self.urlparams = URLParamList(( unquote(s) for s in urlparams.split('/') )) if urlparams else URLParamList()


    def __repr__(self):
        return 'RoutingData: app={}, page={}, module={}, function={}, urlparams={}'.format(
             self.app,
             self.page,
             self.module if self.class_obj is None else (self.module + '.' + self.class_obj.name),
             self.function,
             self.urlparams,
        )

    def _debug(self):
        return 'django_mako_plus RoutingData:' + \
            ''.join(('\n\t{: <16}{}'.format(k, v) for k, v in (
                ( 'app', self.app ),
                ( 'page', self.page ),
                ( 'module', self.module ),
                ( 'function', self.function ),
                ( 'class_obj', self.class_obj ),
                ( 'function_obj', self.function_obj ),
                ( 'urlparams', self.urlparams ),
            )))


    def render(self, template, context=None, def_name=None, subdir='templates', content_type=None, status=None, charset=None):
        '''App-specific render function that renders templates in the *current app*, attached to the request for convenience'''
        ensure_dmp_app(self.app)
        template_loader = get_dmp_instance().get_template_loader(self.app, subdir)
        template_adapter = template_loader.get_template(template)
        return getattr(template_adapter, 'render_to_response')(context=context, request=self.request, def_name=def_name, content_type=content_type, status=status, charset=charset)


    def render_to_string(self, template, context=None, def_name=None, subdir='templates'):
        '''App-specific render function that renders templates in the *current app*, attached to the request for convenience'''
        ensure_dmp_app(self.app)
        template_loader = get_dmp_instance().get_template_loader(self.app, subdir)
        template_adapter = template_loader.get_template(template)
        return getattr(template_adapter, 'render')(context=context, request=request, def_name=def_name)




# This singleton is set on the request object early in the request (during middleware).
# Once urls.py has processed, request.dmp is changed to a populated RoutingData object.
INITIAL_ROUTING_DATA = RoutingData()



################################################################
###   A set of request processors that add variables to the
###   context (parameters) when templates are rendered.
###

from django.conf import settings as conf_settings
from django.template.backends.utils import csrf_input_lazy, csrf_token_lazy



def settings(request):
    '''Adds the settings dictionary to the request'''
    return { 'settings': conf_settings }


def csrf(request):
    '''
    Adds the "csrf_input" and "csrf_token" variables to the request.

    Following Django's lead, this processor is included in DMP's
    default context processors list. It does not need to be listed
    in settings.py.

    To include the <input name="csrf".../> control in your forms,
    use ${ csrf_input }.
    '''
    return {
        'csrf_input': csrf_input_lazy(request),
        'csrf_token': csrf_token_lazy(request),
    }
